/*
 * Copyright (C) 1997-2005, R3vis Corporation.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA,
 * or visit http://www.gnu.org/copyleft/gpl.html.
 *
 * Contributor(s):
 *   Wes Bethel, R3vis Corporation, Marin County, California
 *
 * The OpenRM project is located at http://openrm.sourceforge.net/.
 */
/*
 * $Id: pntcld.c,v 1.14 2005/06/08 18:34:16 wes Exp $
 * $Revision: 1.14 $
 * $Name: OpenRM-1-6-0-2-c $
 * $Log: pntcld.c,v $
 * Revision 1.14  2005/06/08 18:34:16  wes
 * Code cleanup to eliminate compiler warnings.
 *
 * Revision 1.13  2005/05/16 01:16:58  wes
 * Add code to exercise texturing on RM_POINTS primitives.
 *
 * Revision 1.12  2003/04/13 18:13:23  wes
 * Updated copyright dates.
 *
 * Revision 1.11  2003/01/27 05:07:07  wes
 * Changes to RMpipe initialization sequence APIs. Tested for GLX, but not WGL.
 *
 * Revision 1.10  2003/01/16 22:22:45  wes
 * Updated all source files to reflect new organization of header files: all
 * headers that were formerly located in include/rmaux, include/rmv and
 * include/rmi are now located in include/rm.
 *
 * Revision 1.9  2002/06/17 00:41:51  wes
 * Added -s command line flag to generate per-frame rendering statistics.
 * replaced rmSubtreeFrame with rmFrame.
 *
 * Revision 1.8  2001/07/15 22:33:19  wes
 * Added rmPipeDelete to the end of all demo progs. For those that use
 * an initfunc, added a new RMnode * parm (which is unused, except for rm2screen).
 *
 * Revision 1.7  2001/06/03 19:44:05  wes
 * Add calls to new rmaux routines to handle window resize events, and
 * for keyboard event handling.
 *
 * Revision 1.6  2001/03/31 16:55:18  wes
 * Added procmode.h, which defines an RMpipe processing mode used in
 * most demonstration programs. The default processing mode is
 * RM_PIPE_MULTISTAGE_VIEW_PARALLEL.
 *
 * Revision 1.5  2000/12/02 17:24:32  wes
 * Version 1.4.0-alpha-1 checkin. See the RELEASENOTES file for
 * a summary of changes. With this checkin, these demo programs
 * are no longer compatible with versions of the OpenRM API that
 * are pre-1.4.0.
 *
 * Revision 1.4  2000/08/28 01:38:18  wes
 * Updated rmaux* interfaces - rmauxEventLoop now takes two additional
 * parameters (keypress and window resize app callbacks); replaced
 * rmauxUI with rmauxSetGeomTransform and, where appropriate,
 * rmauxSetCamera3DTransform.
 *
 * Revision 1.3  2000/04/27 03:12:05  wes
 * Minor code cleanup.
 *
 * Revision 1.2  2000/04/20 18:04:34  wes
 * Minor tweaks and reorg for OpenRM 1.2.1.
 *
 * Revision 1.1.1.1  2000/02/28 21:55:30  wes
 * OpenRM 1.2 Release
 *
 * Revision 1.11  2000/02/28 17:21:55  wes
 * RM 1.2, pre-OpenRM
 *
 */

#include <rm/rm.h>   
#include <rm/rmaux.h>
#include <rm/rmi.h>
#include <time.h>
#include "procmode.h"

/*
 * usage: pointcloud [-w img_width] [-h img_height] [-n num_points] [-s (display framerate stats)] [-tc (use texture mapping; generate tc's for each point, and use a default image for the texture)]
 */

static char MyRootName[]={"MyRoot"};
static RMnode *MyRoot;
static int img_width=400,img_height=300;
static int num_points=2500;
static int do_pointsize=0;
static float pointsize=0.0;
static int doStats=0;
static int doTexture=0;

void
usage(char *av[])
{
    char buf[256];
    sprintf(buf," usage: %s [-w img_width] [-h img_height] [-n num_points] [-p size (point size in pixels) ] [-s (turn on compute/report of frame rate stats)] [-tc (use texture mapping; generate tc's for each point, and use a default image for the texture)] \n",av[0]);
    rmError(buf);
}

void
parse_args(int ac,
	   char *av[])
{
    int i;

    i = 1;
    while (i < ac)
    {
	if (strcmp(av[i],"-w") == 0)
	{
	    i++;
	    sscanf(av[i],"%d",&img_width);
	}
	else if (strcmp(av[i],"-h") == 0)
        {
	    i++;
	    sscanf(av[i],"%d",&img_height);
	}
	else if (strcmp(av[i],"-s") == 0)
	{
	    i++;
	    doStats = 1;
	}
	else if (strcmp(av[i],"-n") == 0)
	{
	    i++;
	    sscanf(av[i],"%d",&num_points);
	}
	else if (strcmp(av[i],"-p") == 0)
	{
	    i++;
	    do_pointsize = 1;
	    sscanf(av[i],"%f",&pointsize);
	}
	else if (strcmp(av[i],"-tc") == 0)
	{
	    doTexture = 1;
	}
	else
	{
	    usage(av);
	    exit(-1);
	}
	i++;
    }
}

void
my_set_scene(RMnode *camNode,
	     RMenum stereo_format)
{
    /*
     * here, we compute the camera parameters such that all of
     * the geometry is visible.
     */
      
    RMcamera3D *c=rmCamera3DNew();

    /* assign a default view */
    rmDefaultCamera3D(c);

    /* adjust the default view so that all geom is visible */
    rmCamera3DComputeViewFromGeometry(c,MyRoot,img_width,img_height);

    if (stereo_format != RM_MONO_CHANNEL)
    {
	rmCamera3DSetStereo(c,RM_TRUE);
	rmCamera3DSetEyeSeparation(c,2.5F);
	rmCamera3DSetFocalDistance (c,0.707F);

    }
    rmNodeSetSceneCamera3D(camNode, c); 
    rmCamera3DDelete(c);

    /* use RM's default lighting model */
    rmDefaultLighting(camNode);
}

#define DIVISOR 1./(double)RAND_MAX

void
ToVertex(int r1,
	 int r2,
	 RMvertex3D *v,
	 RMvertex3D *n,
	 RMvertex2D *tc)
{
    double azimuth, elevation;
    double ca,sa,se,ce;

    /* convert to range 0..1 */
    azimuth = (double)r1 * DIVISOR;
    elevation = (double)r2 * DIVISOR;

    tc->x = azimuth;
    tc->y = elevation;

    /* convert azimuth to range 0..2pi */
    azimuth *= RM_PI * 2.0;
    
    /* convert elevation to range -pi/2..pi/2 */
    elevation = (elevation - 0.5) * RM_PI;

    /* covert  */

    
    ca = cos(azimuth);
    sa = sin(azimuth);
    se = sin(elevation);
    ce = cos(elevation);

    v->y = se;
    v->x = ca * ce;
    v->z = sa * ce;

    n->x = v->x;
    n->y = v->y;
    n->z = v->z;
}

void
loadTexture(RMnode *addTo)
{
    RMimage *img = rmiReadJPEG("data/seaSurfaceTemperature.jpg");
    RMtexture *t = rmTextureNew(2);

    rmImageMirror(img, RM_IMAGE_MIRROR_HEIGHT);
    rmImageMirror(img, RM_IMAGE_MIRROR_WIDTH);
    rmTextureSetImages(t, &img, 1, 0);
    rmTextureSetEnv(t, GL_MODULATE, NULL);
    rmNodeSetSceneTexture(addTo, t);

    rmTextureDelete(t, RM_TRUE);
}

void
my_build_objs(void)
{
    RMnode *objnode;

    MyRoot = rmNodeNew(MyRootName,RM_RENDERPASS_3D, RM_RENDERPASS_OPAQUE);
    rmNodeAddChild(rmRootNode(),MyRoot);
    
    {
	RMprimitive *p;
	int npts = num_points;
	RMvertex3D *v,*n;
	RMvertex2D *tc;
	RMcolor3D *c;
	int i;
	
	objnode = rmNodeNew("pointcloud",RM_RENDERPASS_3D, RM_RENDERPASS_OPAQUE);
	p = rmPrimitiveNew(RM_POINTS);

	v = rmVertex3DNew(npts);
	n = rmVertex3DNew(npts);
	tc = rmVertex2DNew(npts);
	c = (RMcolor3D *)rmVertex3DNew(npts);

	srand(time(NULL));
	/* generate random points on a the surface of a sphere */
	for (i=0;i<npts;i++)
	{
	    int r1,r2;
	    r1 = rand();
	    r2 = rand();
	    ToVertex(r1,r2,v+i,n+i, tc+i);

	    c[i].r = c[i].g = c[i].b = 0.5;
	    c[i].r = v[i].y * 0.5 + 0.5;
	}


	rmPrimitiveSetVertex3D(p,npts,v,RM_COPY_DATA,NULL);
	rmPrimitiveSetColor3D(p, npts, c, RM_COPY_DATA, NULL);
	rmPrimitiveSetNormal3D(p,npts,n, RM_COPY_DATA, NULL);

	if (doTexture)
	    rmPrimitiveSetTexcoord2D(p, npts, tc, RM_COPY_DATA, NULL);

	rmNodeAddPrimitive (objnode, p);

	rmVertex3DDelete(v);
	rmVertex3DDelete(n);
	rmVertex2DDelete(tc);

	rmNodeComputeBoundingBox(objnode);
	rmNodeAddChild(MyRoot, objnode);
	
	if (doTexture == 1)
	    loadTexture(objnode);
    }

    rmNodeUnionAllBoxes(MyRoot);
    rmNodeComputeCenterFromBoundingBox(MyRoot);
    
    {
	RMcolor4D bgcolor={0.2,0.2,0.3,1.0};
	/*
	 * assign a background color to take effect at "MyRoot"
	 */
	rmNodeSetSceneBackgroundColor(MyRoot,&bgcolor);
    }

    if (do_pointsize)
	rmNodeSetPointSize(MyRoot,pointsize);
}

void
my_idle_func(RMpipe *p,
	     int ix,
	     int iy)
{
    RMmatrix m,old;
    double d,c,s;
    rmMatrixIdentity(&m);
    d = RM_DEGREES_TO_RADIANS(1.0);
    c = cos(d);
    s = sin(d);
    m.m[0][0] = m.m[2][2] = c;
    m.m[0][2] = -s;
    m.m[2][0] = s;
    
    rmNodeGetRotateMatrix(MyRoot,&old);
    rmMatrixMultiply(&old,&m,&old);
    rmNodeSetRotateMatrix(MyRoot,&old);

    rmFrame(p, rmRootNode());

    /* foil compiler warning */
    ix = iy = 0;
}

void
myInitFunc(RMpipe *p,
	   RMnode *n)
{
    my_build_objs();
    my_set_scene(n, rmPipeGetChannelFormat(p));
    rmauxSetGeomTransform(MyRoot,p);
    rmauxSetCamera3DTransform(n, p);

    /*
     * set handler to reset aspect ratio when the window is resized.
     */
    rmauxSetResizeFunc(p, n, rmauxDefaultResizeFunc);

    if (doStats != 0)
	rmStatsComputeDemography(n);
    
    if (rmPipeProcessingModeIsMultithreaded(p) == RM_TRUE)
	rmFrame(p, n);
    
    rmFrame(p, n);
}

void
myRenderFunc(RMpipe *p,
	     RMnode *n)
{
    if (doStats != 0)
	rmStatsStartTime();
    
    rmFrame(p, n);

    if (doStats != 0)
    {
	rmStatsEndTime();
	rmStatsPrint();
    }
}
#ifdef RM_WIN
int WINAPI WinMain (HINSTANCE hInstance,
		    HINSTANCE hPrevInstance,
                    LPSTR lpszCmdLine, int nCmdShow)
{
    MSG      msg; 
    HWND     hWnd; 
    void *fptr;
    RMenum channelFormat;
    RMenum processingMode = DEFAULT_PROCESSING_MODE; /* in procmode.h */
    RMpipe *lone_pipe=NULL;
    RMenum targetPlatform = RM_PIPE_WGL;

    parse_args(__argc,__argv);
#else
int
main(int ac,
     char *av[])
{
    RMpipe *lone_pipe=NULL;
    RMenum processingMode = DEFAULT_PROCESSING_MODE; /* in procmode.h */
    RMenum channelFormat;
    RMenum targetPlatform = RM_PIPE_GLX;
    
    void *msg;			/* needed for rmauxEventLoop
				 win32/unix API consistency */

    parse_args(ac,av);
#endif
    /* 
     * pick a stereo format:
     * RM_MONO_CHANNEL - plain old single-view
     * RM_REDBLUE_STEREO_CHANNEL - left channel in red, right channel in cyan
     * RM_BLUERED_STEREO_CHANNEL - left in cyan, right in red
     * RM_MBUF_STEREO_CHANNEL - multibuffered stereo, requires special
     *    hardware. 
     */
    channelFormat = RM_MONO_CHANNEL; 

    /* 
     * first stage of RM initialization.
     */
    rmInit();

    /* 
     * create the rendering pipe. this step is required in both
     * Win32 and X.
     */
    lone_pipe = rmPipeNew(targetPlatform);
    
    rmPipeSetProcessingMode(lone_pipe, processingMode);

#ifdef RM_WIN
    {
        /*
	 * Win32: when a window is created, we have to tell windows the
	 * name of the "WndProc," the procedure that gets called by
	 * windows with events (the event loop) (contrast to the X model 
	 * where the name of the event loop is not part of the window). 
	 * Since we're using RMaux, we know about the event handling 
	 * procedure named "rmauxWndProc" and we provide that here. 
	 */

        fptr = (void *)(rmauxWndProc);
	hWnd = rmauxCreateW32Window(lone_pipe,
			       NULL, /* no parent window */
			       20,20,img_width,img_height,"RM for Windows",
			       hInstance,fptr);
	if (hWnd == 0)
	  exit(-1);

	/* 
	 * assign the new window handle to the rendering pipe.
	 */
	rmPipeSetWindow(lone_pipe,hWnd, img_width, img_height);
    }
#endif
    
#ifdef RM_X
    {
	Window w;

	w = rmauxCreateXWindow(lone_pipe,
			       (Window)NULL, /* parent window */
			       0,0,img_width,img_height,
			       "RM for X-Windows","icon-title",RM_TRUE);
	/* 
	 * assign the window to the rendering pipe.
	 */
	rmPipeSetWindow(lone_pipe,w,img_width,img_height);

    }
#endif

    /* 
     * specify the name of the "init" function. the "init" function is
     * mandatory in the Win32 world, and optional in the X world. 
     *
     * in Win32, we don't want to call RM services until OpenGL is
     * ready. we can be assured of readiness by using an init function
     * with RMaux. 
     *
     * in X, at this point, the window is mapped and OpenGL is ready,
     * and we could call our init function directly.
     */

    rmauxSetInitFunc(myInitFunc); 


    /* uncomment this next line if you want the object to rotate
       while the user is idle. */
    /*    rmauxSetIdleFunc(lone_pipe, my_idle_func); */
    
    /* 
     * X-ism: once the window is created and assigned to the 
     * rendering pipe, rmUsePipe makes the OpenGL rendering context
     * current for the pipe+window combination. 
     *
     * this step is required for X. in these demo programs, it is not 
     * strictly required by Win32, as the newly created context is made
     * current as part of the OpenGL initialization sequence.
     */
    rmPipeMakeCurrent(lone_pipe);	

    rmauxSetRenderFunc(myRenderFunc);
    /*
     * set key handler function so this prog will exit on "q" key.
     */
    rmauxSetKeyFunc(lone_pipe, rmauxDefaultKeyFunc);
    
    rmauxEventLoop(lone_pipe, rmRootNode(), &msg);

    rmPipeDelete(lone_pipe);
    rmFinish();

#ifdef RM_WIN
    return( msg.wParam );
#else
    return(1);
#endif

}
