/*
 * Copyright (C) 1997-2005, R3vis Corporation.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA,
 * or visit http://www.gnu.org/copyleft/gpl.html.
 *
 * Contributor(s):
 *   Wes Bethel, R3vis Corporation, Marin County, California
 *
 * The OpenRM project is located at http://openrm.sourceforge.net/.
 */
/*
 * $Id: vector3d.c,v 1.12 2005/06/08 15:23:08 wes Exp $
 * $Revision: 1.12 $
 * $Name: OpenRM-1-6-0-2-c $
 * $Log: vector3d.c,v $
 * Revision 1.12  2005/06/08 15:23:08  wes
 * Clean up compiler warnings. Change image output to JPEG rather than AVS.
 *
 * Revision 1.11  2003/04/13 18:13:23  wes
 * Updated copyright dates.
 *
 * Revision 1.10  2003/01/27 05:07:07  wes
 * Changes to RMpipe initialization sequence APIs. Tested for GLX, but not WGL.
 *
 * Revision 1.9  2003/01/16 22:22:45  wes
 * Updated all source files to reflect new organization of header files: all
 * headers that were formerly located in include/rmaux, include/rmv and
 * include/rmi are now located in include/rm.
 *
 * Revision 1.8  2002/06/17 00:46:03  wes
 * Replaced rmSubtreeFrame with rmFrame.
 *
 * Revision 1.7  2001/07/15 22:33:19  wes
 * Added rmPipeDelete to the end of all demo progs. For those that use
 * an initfunc, added a new RMnode * parm (which is unused, except for rm2screen).
 *
 * Revision 1.6  2001/06/03 19:44:05  wes
 * Add calls to new rmaux routines to handle window resize events, and
 * for keyboard event handling.
 *
 * Revision 1.5  2001/03/31 16:55:18  wes
 * Added procmode.h, which defines an RMpipe processing mode used in
 * most demonstration programs. The default processing mode is
 * RM_PIPE_MULTISTAGE_VIEW_PARALLEL.
 *
 * Revision 1.4  2000/12/02 17:24:32  wes
 * Version 1.4.0-alpha-1 checkin. See the RELEASENOTES file for
 * a summary of changes. With this checkin, these demo programs
 * are no longer compatible with versions of the OpenRM API that
 * are pre-1.4.0.
 *
 * Revision 1.3  2000/08/28 01:38:18  wes
 * Updated rmaux* interfaces - rmauxEventLoop now takes two additional
 * parameters (keypress and window resize app callbacks); replaced
 * rmauxUI with rmauxSetGeomTransform and, where appropriate,
 * rmauxSetCamera3DTransform.
 *
 * Revision 1.2  2000/04/20 18:04:34  wes
 * Minor tweaks and reorg for OpenRM 1.2.1.
 *
 * Revision 1.1.1.1  2000/02/28 21:55:30  wes
 * OpenRM 1.2 Release
 *
 * Revision 1.11  2000/02/28 17:21:56  wes
 * RM 1.2, pre-OpenRM
 *
 */
#include <rm/rm.h>
#include <rm/rmaux.h>
#include <rm/rmv.h>
#include <rm/rmi.h>
#include "libdio.h"
#include "procmode.h"

static RMpipe *lone_pipe=NULL;
static char MyRootName[]={"MyRoot"};
static RMnode *MyRoot;
int img_width=400,img_height=300;
char datafilename[256]={"data/func10.dio"};
dioDataObject *mydataobj=NULL;
RMvertex3D *gradients=NULL;
int do_color=0;
int vis_technique=0;
int my_linewidth = RM_LINEWIDTH_MEDIUM;
int my_linestyle = RM_LINES_SOLID;
int scale_type = RMV_NO_SCALE;
float scale_value = 1.0;


void
usage(char *av[])
{
    fprintf(stderr," usage: %s [-i datafilename (defaults to data/func10.dio) [-w img_width] [-h img_height] [-c (use default vis colormap to 'colorize' the plot') [-v n (where n=0..2, and indicates which visualization technique to use)]  [-st n (set the vector scale type, where n = 0..2 for RMV_NO_SCALE, RMV_LINEAR_SCALE or RMV_LOG10_SCALE)] [-sv xx.x (where xx.x is a floating point value to use for vector scaling according to the scale type)] \n",av[0]);
}

void
parse_args(int ac,
	   char *av[])
{
    int i;

    i = 1;
    while (i < ac)
    {
	if (strcmp(av[i],"-w") == 0)
	{
	    i++;
	    sscanf(av[i],"%d",&img_width);
	}
	else if (strcmp(av[i],"-h") == 0)
        {
	    i++;
	    sscanf(av[i],"%d",&img_height);
	}
	else if (strcmp(av[i],"-i") == 0)
	{
	    i++;
	    strcpy(datafilename,av[i]);
	}
	else if (strcmp(av[i],"-c") == 0)
	{
	    do_color=1;
	}
	else if (strcmp(av[i],"-v") == 0)
	{
	    i++;
	    sscanf(av[i],"%d",&vis_technique);

            if ((vis_technique < 0) || (vis_technique > 2))
            {
                usage(av);
                exit(0);
            }
	}
	else if (strcmp(av[i],"-st") == 0)
	{
	    int j;
	    i++;
	    sscanf(av[i],"%d",&j);
	    switch (j)
	    {
	    case 0:
		scale_type = RMV_NO_SCALE;
		break;
	    case 1:
		scale_type = RMV_LINEAR_SCALE;
		break;
	    case 2:
		scale_type = RMV_LOG10_SCALE;
		break;
	    default:
		fprintf(stderr," error: specify a scale type with -st 0 for RMV_NO_SCALE, -st 1 for RMV_LINEAR_SCALE or -st 2 for RMV_LOG10_SCALE. bye! \n");
		exit(-1);
	    }
	}
	else if (strcmp(av[i],"-sv") == 0)
	{
	    i++;
	    sscanf(av[i],"%f",&scale_value);
	}
	else
	{
	    usage(av);
	    exit(-1);
	}
	i++;
    }
}

void
my_read_data(char *datafilename)
{
    mydataobj = dioReadDataObject(datafilename);
    if (mydataobj == NULL)
    {
	fprintf(stderr," error reading input data file. exiting. \n");
	exit(-1);
    }
    dioObjectConditioner(mydataobj);
}

void
my_set_scene(RMnode *camNode,
	     int stereo_format)
{
      
    RMcamera3D *c=rmCamera3DNew();
    
    /* create a camera */
    rmDefaultCamera3D(c);		/* assign it some default values. */

    /* adjust default view so all geom is visible */
    rmCamera3DComputeViewFromGeometry(c,MyRoot, img_width, img_height);

    if (stereo_format != RM_MONO_CHANNEL)
    {
	rmCamera3DSetStereo(c,RM_TRUE);
	rmCamera3DSetEyeSeparation(c,2.5F);
	rmCamera3DSetFocalDistance (c,0.707F);
	
    }

    /* add the camera to "my root's" scene parms. */
    rmNodeSetSceneCamera3D(camNode, c);

    rmCamera3DDelete(c);
    
    /* use RM's default lighting model */
    rmDefaultLighting(camNode);
}

void
my_compute_gradients(float *xcoords,
		     float *ycoords,
		     float *zcoords,
		     int usize,
		     int vsize,
		     RMvertex3D *ret_normals)
{
    int i,j;
    int indx = 0;

    for (j=0;j<vsize;j++)
    {
	int plus_v,minus_v;

	if (j == 0)
	    minus_v = 0;
	else
	    minus_v = -usize;

	if (j == vsize-1)
	    plus_v = 0;
	else
	    plus_v = usize;
	
	for (i=0;i<usize;i++)
	{
	    RMvertex3D p,r,c;
	    int plus_u,minus_u;

	    /* at each data point, compute the data normal */

	    if (i == 0)
		minus_u = 0;
	    else
		minus_u = -1;

	    if (i == usize-1)
		plus_u = 0;
	    else
		plus_u = 1;

	    p.x = xcoords[j*usize+i + plus_u] - xcoords[j*usize+i + minus_u];
	    p.y = ycoords[j*usize+i + plus_u] - ycoords[j*usize+i + minus_u];
	    p.z = zcoords[j*usize+i + plus_u] - zcoords[j*usize+i + minus_u];
	    
	    r.x = xcoords[j*usize+i + plus_v] - xcoords[j*usize+i + minus_v];
	    r.y = ycoords[j*usize+i + plus_v] - ycoords[j*usize+i + minus_v];
	    r.z = zcoords[j*usize+i + plus_v] - zcoords[j*usize+i + minus_v];

	    rmVertex3DCross(&p,&r,&c);

	    ret_normals[indx] = c;

	    indx++;
	}
    }
}

/*
 * the following two routines are the interface between the RMV
 * vis tools and the local data model. RMV wants us to supply routines
 * which will tell the vis tool what the (x,y,z) point is at some grid
 * location, and what the data point is at some grid location.
 *
 * the local data model is very simple, so we can make simplifying
 * assumptions resulting in very terse routines.
 */
RMvertex3D
mygridfunc_u(int i)
{
    /*
     * tell RMV what this grid (x,y,z) point is at location "i". we assume
     * a one-d grid of (x,y,z) points in the local data model.
     */
    RMvertex3D temp3d;
    
    temp3d.x = mydataobj->xcoords[i];
    temp3d.y = mydataobj->ycoords[i];
    temp3d.z = mydataobj->zcoords[i];
    return(temp3d);
}

float
mydatafunc_u(int i)
{
    /*
     * tell RMV what the data value is at grid location "i". we
     * assume a one-d grid.
     */
    return(mydataobj->rawdata[i]);
}

RMvertex3D
myvectordatafunc_u(int i)
{
    return(gradients[i]);
}

void
my_build_objs(void)
{
    MyRoot = rmNodeNew(MyRootName,RM_RENDERPASS_3D, RM_RENDERPASS_OPAQUE);
    rmNodeAddChild(rmRootNode(),MyRoot);
    
    /* do the visualization.. */
    {
	RMnode *visnode, *vectorvisnode;
	int offset_flag;
	float *xcoords,*ycoords,*zcoords, *data,*data2;
	int usize,vsize;
	RMvisMap *vmap=NULL;

	/* we assume that one and only one of w,h,d is equal to 1 */
	
	if (mydataobj->height == 1)
	{
	    /* width & depth are != 1, so offset from the y axis */
	    offset_flag = RMV_YAXIS_OFFSET;
	    usize = mydataobj->width;
	    vsize = mydataobj->depth;
	}
	else if (mydataobj->width == 1)
	{
	    /* height & depth are != 1, so offset from the X axis */
	    offset_flag = RMV_XAXIS_OFFSET;
	    usize = mydataobj->height;
	    vsize = mydataobj->depth;
	}
	else			/* assume depth==1, offset from z axis */
	{
	    offset_flag = RMV_ZAXIS_OFFSET;
	    usize = mydataobj->width;
	    vsize = mydataobj->height;
	}

	if (do_color)
	{
	    vmap = rmDefaultVismap();

	    rmVismapSetTfMin(vmap,mydataobj->datamin);
	    rmVismapSetTfMax(vmap,mydataobj->datamax);
	    data2 = mydataobj->rawdata;
	}

	visnode = rmNodeNew("vis",RM_RENDERPASS_3D, RM_RENDERPASS_ALL);
	vectorvisnode = rmNodeNew("vectorvis",RM_RENDERPASS_3D, RM_RENDERPASS_OPAQUE);

	xcoords = mydataobj->xcoords;
	ycoords = mydataobj->ycoords;
zcoords = mydataobj->zcoords;
	data = mydataobj->rawdata;

	/*
	 * compute 3d gradients from the source data for use in
	 * demonstrating the vector visualization techniques. we're going
	 * to assume that the input data is to be offset from the z axis.
	 */

	gradients = rmVertex3DNew(usize*vsize);
	
	my_compute_gradients(xcoords,ycoords,data, usize,vsize,
			     gradients);


	switch(vis_technique)
	{
	case 0:
	    rmvI3VectorBarbs(mygridfunc_u,
			     mydatafunc_u,
			     (vmap == NULL) ? NULL : mydatafunc_u,
			     myvectordatafunc_u,
			     vmap, 
			     offset_flag, usize*vsize,
			     scale_value, scale_type,
			     my_linewidth, my_linestyle,
			     visnode);
	    break;

	case 1:
	    rmvI3VectorArrows(mygridfunc_u,
			      mydatafunc_u,
			      (vmap == NULL) ? NULL : mydatafunc_u,
			      myvectordatafunc_u,
			      vmap, 
			      offset_flag, usize*vsize,
			      scale_value,  scale_type,
			      my_linewidth, my_linestyle,
			      visnode);
	    break;

	case 2:
	    rmvI3VectorCones(mygridfunc_u,
			     mydatafunc_u,
			     (vmap == NULL) ? NULL : mydatafunc_u,
			     myvectordatafunc_u,
			     vmap, 
			     offset_flag, usize*vsize,
			     scale_value,  scale_type,
			     visnode);
	    break;

	default: /* bogus vector viz technique */
	    break;
	}

	rmNodeAddChild(MyRoot,visnode);
	rmNodeComputeBoundingBox(visnode);
	
	rmNodeUnionAllBoxes(MyRoot);
	rmNodeComputeCenterFromBoundingBox(MyRoot);
    }

    {
	RMcolor4D bgcolor={0.2,0.2,0.3,1.0};
	rmNodeSetSceneBackgroundColor(MyRoot, &bgcolor); 
    }
}

void
my_idle_func(RMpipe *p,
	     int ix, 
	     int iy)
{
    RMmatrix m,old;
    double d,c,s;
    rmMatrixIdentity(&m);
    d = RM_DEGREES_TO_RADIANS(1.0);
    c = cos(d);
    s = sin(d);
    m.m[0][0] = m.m[2][2] = c;
    m.m[0][2] = -s;
    m.m[2][0] = s;
    
    if (rmNodeGetRotateMatrix(MyRoot,&old) == RM_WHACKED)
	rmMatrixIdentity(&old);
    rmMatrixMultiply(&old,&m,&old);
    rmNodeSetRotateMatrix(MyRoot,&old);

    /* foil compiler warnings */
    ix = iy = 0;

    rmFrame(p, rmRootNode());
}

void
dumpimagefunc(const RMimage *img,
	      RMenum whichbufferEnum)
{
    /* foil compiler warning */
    whichbufferEnum = RM_TRUE;

    rmiWriteJPEG("vector2d.jpg", 99, img);
}

int
my_dump_image_func(RMpipe *p,
		   int xbutton,
		   int ybutton)
{
    /*
     * the goal is to write an image file that contains the
     * contents of the framebuffer.
     */
    rmNotice(" in my_dump_image_func(). \n");
    
    /*
     * assign a "post render" function - it will be invoked after
     * the scene has been rendered, and will write the framebuffer
     * contents to a file.
     */
    rmPipeSetPostRenderFunc(lone_pipe,dumpimagefunc);
    
    /*
     * render & write the image.
     */
    rmFrame(p, rmRootNode());

    /*
     * now, remove the "write image" function from the pipe. if we
     * didn't remove this callback, we'd write a file ever time the
     * frame is rendered - which is useful, but not what we want
     * in this context.
     */
    rmPipeSetPostRenderFunc(lone_pipe,NULL);

    /* foil compiler warning */
    xbutton = ybutton = 0;

    return(1);
}

void
myinitfunc(RMpipe *p,
	   RMnode *n)
{
    my_read_data(datafilename);
    my_build_objs();
    my_set_scene(n, rmPipeGetChannelFormat(p));
    rmauxSetGeomTransform(MyRoot, p);
    rmauxSetCamera3DTransform(n, p);

    rmauxSetButtonDownFunc(RM_BUTTON1,RM_CONTROL_MODMASK,my_dump_image_func);
    /*
     * set handler to reset aspect ratio when the window is resized.
     */
    rmauxSetResizeFunc(p, rmRootNode(), rmauxDefaultResizeFunc);

    if (rmPipeProcessingModeIsMultithreaded(p) == RM_TRUE)
	rmFrame(p, n);
    
    rmFrame(p, n);
}

#ifdef RM_WIN
int WINAPI WinMain (HINSTANCE hInstance,
		    HINSTANCE hPrevInstance,
                    LPSTR lpszCmdLine, int nCmdShow)
{
    MSG      msg; 
    HWND     hWnd; 
    void *fptr;
    RMenum channelFormat;
    RMenum processingMode = DEFAULT_PROCESSING_MODE; /* in procmode.h */
    RMenum targetPlatform = RM_PIPE_WGL;
    RMpipe *lone_pipe=NULL;

    parse_args(__argc, __argv);
#else  /* assume RM_X */
int
main(int ac,
     char *av[])
{

    RMpipe *lone_pipe=NULL;
    RMenum processingMode = DEFAULT_PROCESSING_MODE; /* in procmode.h */
    RMenum channelFormat;
    RMenum targetPlatform = RM_PIPE_GLX;
    void *msg;			/* needed for rmauxEventLoop
				 win32/unix API consistency */

    parse_args(ac,av);
#endif

    /* 
     * pick a stereo format:
     * RM_MONO_CHANNEL - plain old single-view
     * RM_REDBLUE_STEREO_CHANNEL - left channel in red, right channel in cyan
     * RM_BLUERED_STEREO_CHANNEL - left in cyan, right in red
     * RM_MBUF_STEREO_CHANNEL - multibuffered stereo, requires special
     *    hardware. 
     */
    channelFormat = RM_BLUERED_STEREO_CHANNEL; /* RM_MONO_CHANNEL */


    /* 
     * first stage of RM initialization.
     */
    rmInit();

    /* 
     * create the rendering pipe. this step is required in both
     * Win32 and X.
     */
    lone_pipe = rmPipeNew(targetPlatform);

    rmPipeSetChannelFormat(lone_pipe, channelFormat);

    rmPipeSetProcessingMode(lone_pipe, processingMode);
    

#ifdef RM_WIN
    {
        /*
	 * Win32: when a window is created, we have to tell windows the
	 * name of the "WndProc," the procedure that gets called by
	 * windows with events (the event loop) (contrast to the X model 
	 * where the name of the event loop is not part of the window). 
	 * Since we're using RMaux, we know about the event handling 
	 * procedure named "rmauxWndProc" and we provide that here. 
	 */

        fptr = (void *)(rmauxWndProc);
	hWnd = rmauxCreateW32Window(lone_pipe,
			       NULL, /* no parent window */
			20,20,img_width,img_height,"RM for Windows",
			       hInstance,fptr);
	if (hWnd == 0)
	  exit(-1);

	/* 
	 * assign the new window handle to the rendering pipe.
	 */
	rmPipeSetWindow(lone_pipe,hWnd, img_width, img_height);

    }
#endif
#ifdef RM_X
    {
	Window w;

	w = rmauxCreateXWindow(lone_pipe,
			       (Window)NULL, /* parent window */
			       0,0,img_width,img_height,
			       "RM for X-Windows","icon-title",RM_TRUE);
	/* 
	 * assign the window to the rendering pipe.
	 */
	rmPipeSetWindow(lone_pipe,w,img_width,img_height);

    }
#endif

    /* 
     * specify the name of the "init" function. the "init" function is
     * mandatory in the Win32 world, and optional in the X world. 
     *
     * in Win32, we don't want to call RM services until OpenGL is
     * ready. we can be assured of readiness by using an init function
     * with RMaux. 
     *
     * in X, at this point, the window is mapped and OpenGL is ready,
     * and we could call our init function directly.
     */

    rmauxSetInitFunc(myinitfunc); 

    /* uncomment this next line if you want the object to rotate
       while the user is idle. */
    /*    rmauxSetIdleFunc(lone_pipe,my_idle_func); */
    
    /* 
     * X-ism: once the window is created and assigned to the 
     * rendering pipe, rmUsePipe makes the OpenGL rendering context
     * current for the pipe+window combination. 
     *
     * this step is required for X. in these demo programs, it is not 
     * strictly required by Win32, as we made the newly created context
     * current as part of the OpenGL initialization sequence.
     */
    rmPipeMakeCurrent(lone_pipe);	

    /*
     * set key handler function so this prog will exit on "q" key.
     */
    rmauxSetKeyFunc(lone_pipe, rmauxDefaultKeyFunc);
    
    rmauxEventLoop(lone_pipe,rmRootNode(),&msg);
    
    rmPipeDelete(lone_pipe);
    rmFinish();

#ifdef RM_WIN
    return( msg.wParam );
#else
    return(1);
#endif
}
